from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

from os import path, chdir, getcwd, remove
from shutil import rmtree, copy, copytree, move
from tempfile import mkdtemp
from glob import glob
import warnings
import logging
from socket import timeout

import pytest
import numpy as np
from numpy.testing import assert_almost_equal
from astropy.io import fits
from astropy.coordinates import Angle, FK5, name_resolve, SkyCoord
from astropy import units as u
from astropy.table import Table
from astropy.extern import six
from astropy.time import Time

from .. import patchers as ph
from ..feder import Feder, ApogeeAltaU9, FederSite
from ...tests.data import get_data_dir
from ... import ImageFileCollection

pytest_plugins = str("capturelog")

_test_dir = ''
_default_object_file_name = 'obsinfo.txt'
_test_image_name = 'uint16.fit'
simbad_down = False
OBJECT_LIST_URL = 'https://raw.github.com/mwcraig/feder-object-list/master/feder_object_list.csv'


@pytest.mark.usefixtures('object_file_no_ra')
def test_read_object_list():
    objects, ra_dec = ph.read_object_list(directory=_test_dir,
                                          skip_lookup_from_object_name=True)
    assert len(objects) == 2
    assert objects[0] == 'ey uma'
    assert objects[1] == 'm101'
    assert not ra_dec


def test_read_object_list_ra_dec():
    temp_dir = mkdtemp()
    obj_name = 'objects_with_ra.txt'
    object_path = path.join(temp_dir, obj_name)
    object_in = 'ey uma'
    RA_in = "09:02:20.76"
    Dec_in = "+49:49:09.3"
    to_write = 'object, RA, Dec\n{},{},{}'.format(object_in, RA_in, Dec_in)
    object_file = open(object_path, 'wt')
    object_file.write(to_write)
    object_file.close()
    obj, ra_dec = ph.read_object_list(temp_dir, obj_name)
    assert(obj[0] == object_in)
    ra_dec_in = FK5(RA_in, Dec_in, unit=(u.hour, u.degree))
    assert ra_dec_in.separation(ra_dec[0]).arcsec < 1e-4


def test_read_object_list_from_internet():
    try:
        obj, ra_dec = ph.read_object_list(directory='',
                                          input_list=OBJECT_LIST_URL)
    except six.moves.urllib.error.URLError:
        pytest.xfail("Unable to open URL")
    assert 'ey uma' in obj


def test_read_object_list_with_skip_consistency_skip_lookup():
    # this is bad because there is an identical entry
    bad_objects = [
        "ey uma, 09:02:20.76, +49:49:09.3",
        "ey uma, 09:02:20.76, +49:49:09.3"
    ]
    object_file_with_ra_dec(_test_dir, input_objects=bad_objects)
    objects, ra_dec = ph.read_object_list(_test_dir,
                                          skip_lookup_from_object_name=True,
                                          skip_consistency_check=True)
    for obj in objects:
        assert obj == "ey uma"
    assert len(ra_dec) == 2


def test_history_bad_mode():
    with pytest.raises(ValueError):
        ph.history(test_history_bad_mode, mode='not a mode')


def test_history_begin():
    hist = ph.history(test_history_begin, mode='begin')
    assert hist.find('BEGIN') > 0
    assert hist.endswith('+')


def test_history_end():
    hist = ph.history(test_history_end, mode='end')
    assert hist.find('END') > 0
    assert hist.endswith('-')


def test_history_function_name():
    hist = ph.history(test_history_function_name, mode='begin')
    assert hist.find('test_history_function_name') > 0


def test_data_is_unmodified_by_patch_headers():
    """No changes should be made to the data."""
    new_ext = '_new'
    ph.patch_headers(_test_dir, new_file_ext=new_ext)
    test_file_basename = path.splitext(_test_image_name)[0]
    fname = path.join(_test_dir, test_file_basename)
    fname_new = fname + new_ext
    orig = fits.open(fname + '.fit',
                     do_not_scale_image_data=True)
    modified = fits.open(fname_new + '.fit',
                         do_not_scale_image_data=True)
    assert np.all(orig[0].data == modified[0].data)


def test_writing_patched_files_to_directory():
    files = glob(path.join(_test_dir, '*.fit*'))
    n_files_init = len(glob(path.join(_test_dir, '*.fit*')))
    dest_dir = mkdtemp()
    ph.patch_headers(_test_dir, new_file_ext='', save_location=dest_dir)
    print(files)
    n_files_after = len(glob(path.join(_test_dir, '*.fit*')))
    print(n_files_after)
    n_files_destination = len(glob(path.join(dest_dir, '*.fit*')))
    print(dest_dir)
    rmtree(dest_dir)
    assert ((n_files_init == n_files_after) &
            (n_files_init == n_files_destination))


@pytest.mark.parametrize('data_source',
                         ['maximdl_5_21_header.fit',
                          'maximdl_5_23_header.fit'])
def test_purging_maximdl5_keywords(data_source):

    feder = Feder()
    mdl5_name = data_source
    copy(path.join(get_data_dir(), mdl5_name), _test_dir)
    hdr5 = fits.getheader(path.join(_test_dir, mdl5_name))

    software = ph.get_software_name(hdr5, use_observatory=feder)
    ph.purge_bad_keywords(hdr5, history=True, force=False)

    keyword_present = False
    for keyword in software.bad_keywords:
        keyword_present = (keyword_present or (keyword in hdr5))
    assert not keyword_present


@pytest.mark.parametrize('badkey', ['swcreate', 'instrume'])
def test_patch_headers_stops_if_instrument_or_software_not_found(badkey,
                                                                 caplog):
    ic = ImageFileCollection(_test_dir, keywords=['imagetyp'])
    # need a header that contains IMAGETYP so that it will be processed
    a_fits_file = ''
    for h, f in ic.headers(imagetyp='*', return_fname=True):
        a_fits_file = f
        break
    a_fits_hdu = fits.open(path.join(_test_dir, a_fits_file))
    hdr = a_fits_hdu[0].header
    badname = 'Nonsense'
    hdr[badkey] = badname
    a_fits_hdu.writeto(path.join(_test_dir, a_fits_file), clobber=True)
    ph.patch_headers(_test_dir)
    patch_warnings = get_patch_header_logs(caplog)
    assert('KeyError' in patch_warnings)
    assert(badname in patch_warnings)


def test_adding_overscan_apogee_u9(make_overscan_test_files):
    original_dir = getcwd()

    apogee = ApogeeAltaU9()
    print(getcwd())
    oscan_dir, has_oscan, has_no_oscan = make_overscan_test_files
    print(getcwd())

    chdir(path.join(_test_dir, oscan_dir))
    # first, does requesting *not* adding overscan actually leave it alone?
    ph.patch_headers(dir='.', new_file_ext='', overwrite=True, purge_bad=False,
                     add_time=False, add_apparent_pos=False,
                     add_overscan=False, fix_imagetype=False)
    header_no_oscan = fits.getheader(has_no_oscan)
    assert 'biassec' not in header_no_oscan
    assert 'trimsec' not in header_no_oscan

    # Now add overscan
    ph.patch_headers(dir='.', new_file_ext='', overwrite=True, purge_bad=False,
                     add_time=False, add_apparent_pos=False,
                     add_overscan=True, fix_imagetype=False)
    print(_test_dir)
    header_no_oscan = fits.getheader(has_no_oscan)
    # This image had no overscan, so should be missing the relevant keywords.
    assert 'biassec' not in header_no_oscan
    assert 'trimsec' not in header_no_oscan
    header_yes_oscan = fits.getheader(has_oscan)
    # This one as overscan, so should include both of the overscan keywords.
    assert header_yes_oscan['biassec'] == apogee.useful_overscan
    assert header_yes_oscan['trimsec'] == apogee.trim_region
    print(getcwd())
    chdir(original_dir)
    print(getcwd())


def test_fix_imagetype():
    imagetypes_to_check = {'Bias Frame': 'BIAS',
                           'Dark Frame': 'DARK',
                           'Light Frame': 'LIGHT',
                           'Flat Frame': 'FLAT'}
    for im_type in imagetypes_to_check:
        header = fits.Header()

        header['imagetyp'] = im_type
        # first run SHOULD change imagetyp
        print(header)
        ph.change_imagetype_to_IRAF(header, history=False)
        assert(header['imagetyp'] == imagetypes_to_check[im_type])
        # second call should NOT change imagetyp
        print(header)
        ph.change_imagetype_to_IRAF(header, history=True)
        assert(header['imagetyp'] == imagetypes_to_check[im_type])
        with pytest.raises(KeyError):
            print(header['history'])
        # change imagetype back to non-IRAF
        header['imagetyp'] = im_type
        # change with history
        ph.change_imagetype_to_IRAF(header, history=True)
        assert('history' in header)


def test_data_is_unmodified_by_adding_object():
    new_ext = '_obj'
    ph.patch_headers(_test_dir, new_file_ext=new_ext)
    ph.add_object_info(_test_dir, new_file_ext=new_ext)
    test_file_basename = path.splitext(_test_image_name)[0]
    fname = path.join(_test_dir, test_file_basename)
    fname_new = fname + new_ext + new_ext
    orig = fits.open(fname + '.fit',
                     do_not_scale_image_data=True)
    modified = fits.open(fname_new + '.fit',
                         do_not_scale_image_data=True)
    assert np.all(orig[0].data == modified[0].data)


def test_adding_object_name(use_list=None,
                            use_obj_dir=None,
                            check_fits_file=None):
    """
    Test adding object name

    Provide `use_list` to override the default object file name.
    Provide `use_obj_dir` to specify directory in which the object file
    is found. Defaults to directory in which the images reside if
    `use_obj_dir` is None.
    """
    new_ext = '_obj_name_test'
    check_file = check_fits_file or path.splitext(_test_image_name)[0]
    ph.patch_headers(_test_dir, new_file_ext=new_ext)
    ph.add_object_info(_test_dir, new_file_ext=new_ext,
                       object_list=use_list, object_list_dir=use_obj_dir)
    fname = path.join(_test_dir, check_file)
    fname += new_ext + new_ext
    with_name = fits.open(fname + '.fit')
    print('add object name: %s' % fname)
    assert (with_name[0].header['object'] == 'm101')
    return with_name


@pytest.mark.usefixtures('object_file_no_ra')
def test_adding_object_from_name_only():
    if simbad_down:
        pytest.xfail("Simbad is down")
    try:
        test_adding_object_name()
    except (name_resolve.NameResolveError, timeout):
        pytest.xfail("Simbad is down")


@pytest.mark.usefixtures('object_file_ra_change_col_case')
def test_adding_object_name_does_not_depend_on_column_name_case():
    test_adding_object_name()


def test_add_object_name_warns_if_no_match(caplog):
    test_adding_object_name()
    patch_header_warnings = get_patch_header_logs(caplog)
    assert('No object found for image ' in patch_header_warnings)


def test_adding_object_name_to_different_directory(use_list=None,
                                                   use_obj_dir=None):
    new_ext = '_obj_name_test'
    ph.patch_headers(_test_dir, new_file_ext=new_ext)
    destination_dir = mkdtemp()
    ph.add_object_info(_test_dir, new_file_ext=new_ext,
                       save_location=destination_dir,
                       object_list=use_list, object_list_dir=use_obj_dir)
    test_file_basename = path.splitext(_test_image_name)[0]
    fname = path.join(destination_dir, test_file_basename)
    fname += new_ext + new_ext
    with_name = fits.open(fname + '.fit')
    print('add object name: %s' % fname)
    assert (with_name[0].header['object'] == 'm101')
    return with_name


def test_add_object_name_uses_object_list_name():

    custom_object_name = 'my_object_list.txt'
    old_object_path = path.join(_test_dir, _default_object_file_name)
    new_path = path.join(_test_dir, custom_object_name)
    move(old_object_path, new_path)
    fits_with_obj_name = test_adding_object_name(use_list=custom_object_name)
    # The line below is probably not really necessary since the same test
    # is done in test_adding_object_name but it doesn't hurt to test it
    # here too
    assert (fits_with_obj_name[0].header['object'] == 'm101')


def test_add_object_name_with_custom_dir_standard_name():

    a_temp_dir = mkdtemp()

    old_object_path = path.join(_test_dir, _default_object_file_name)
    new_path = path.join(a_temp_dir, _default_object_file_name)
    move(old_object_path, new_path)
    test_adding_object_name(use_obj_dir=a_temp_dir)


def test_add_object_name_uses_object_list_dir():

    a_temp_dir = mkdtemp()

    custom_object_name = 'my_object_list.txt'
    old_object_path = path.join(_test_dir, _default_object_file_name)
    new_path = path.join(a_temp_dir, custom_object_name)
    move(old_object_path, new_path)
    # first make sure object name isn't added if object list can't be found
    with pytest.raises(IOError):
        test_adding_object_name(use_list=custom_object_name)

    # Now make sure it works when we specify the directory; need to redo
    # setup to clear out files made in pass above
    setup_function(test_add_object_name_uses_object_list_dir)
    test_adding_object_name(use_list=custom_object_name,
                            use_obj_dir=a_temp_dir)


def test_ambiguous_object_file_raises_error():
    a_temp_dir = mkdtemp()
    obj_name = 'bad_list.txt'
    obj_path = path.join(a_temp_dir, obj_name)
    object_in = 'ey uma'
    RA_in = "09:02:20.76"
    Dec_in = "+49:49:09.3"
    to_write = 'object, RA, Dec\n{},{},{}\n'.format(object_in, RA_in, Dec_in)
    to_write += '{},{},{}\n'.format('crap', RA_in, Dec_in)
    object_file = open(obj_path, 'wt')
    object_file.write(to_write)
    object_file.close()
    with pytest.raises(RuntimeError):
        test_adding_object_name(use_list=obj_name, use_obj_dir=a_temp_dir)


def test_missing_object_file_issues_warning(caplog):
    remove(path.join(_test_dir, _default_object_file_name))
    ph.add_object_info(_test_dir)
    patch_header_warnings = get_patch_header_logs(caplog)
    assert 'No object list in directory' in patch_header_warnings


def test_no_object_match_for_image_warning_includes_file_name(caplog):
    remove(path.join(_test_dir, _default_object_file_name))
    to_write = ('# comment 1\n# comment 2\nobject,RA,Dec\n'
                'sz lyn,8:09:35.75,+44:28:17.59')
    object_file = open(path.join(_test_dir, _default_object_file_name), 'wt')
    object_file.write(to_write)
    object_file.close()
    ph.patch_headers(_test_dir, new_file_ext='', overwrite=True)
    ph.add_object_info(_test_dir)
    patch_header_warnings = get_patch_header_logs(caplog)
    assert 'No object found' in patch_header_warnings
    assert _test_image_name in patch_header_warnings


def test_missing_object_column_raises_error():
    object_path = path.join(_test_dir, _default_object_file_name)
    object_table = Table.read(object_path, format='ascii')
    object_table.rename_column('object', 'BADBADBAD')
    object_table.write(object_path, format='ascii')
    with pytest.raises(RuntimeError):
        ph.read_object_list(directory=_test_dir,
                            input_list=_default_object_file_name)


@pytest.mark.usefixtures('object_file_no_ra')
def test_read_object_list_logs_error_if_object_on_list_not_found(caplog):
    if simbad_down:
        pytest.xfail('Simbad is down')
    object_path = path.join(_test_dir, _default_object_file_name)
    object_table = Table.read(object_path, format='ascii',
                              comment='#', delimiter=',')
    object_table.add_row(['not_a_simbad_object'])
    object_table.write(object_path, format='ascii')
    with pytest.raises(name_resolve.NameResolveError):
        ph.read_object_list(_test_dir)
    errs = get_patch_header_logs(caplog, level=logging.ERROR)
    assert 'Unable to do lookup' in errs
    ph.add_object_info(_test_dir)
    errs = get_patch_header_logs(caplog, level=logging.ERROR)
    assert 'Unable to add objects--name resolve error' in errs


def test_add_object_name_logic_when_all_images_have_matching_object(caplog):
    ic = ImageFileCollection(_test_dir, keywords=['imagetyp'])
    for h in ic.headers(imagetyp='light', overwrite=True):
        h['imagetyp'] = 'FLAT'
    ph.add_object_info(_test_dir)
    infos = get_patch_header_logs(caplog, level=logging.INFO)
    assert 'NO OBJECTS MATCHED' in infos


@pytest.mark.parametrize('new_file_ext',
                         ['', None])
def test_add_ra_dec_from_object_name(new_file_ext):
    if simbad_down:
        pytest.xfail("Simbad is down")
    full_path = path.join(_test_dir, _test_image_name)
    f = fits.open(full_path, do_not_scale_image_data=True)
    h = f[0].header
    del h['OBJCTRA']
    del h['OBJCTDEC']
    h['OBJECT'] = 'M101'
    with warnings.catch_warnings():
        ignore_from = 'astropy.io.fits.hdu.hdulist'
        warnings.filterwarnings('ignore', module=ignore_from)
        f.writeto(full_path, clobber=True)
    f.close()

    try:
        ph.add_ra_dec_from_object_name(_test_dir, new_file_ext=new_file_ext,
                                       object_list_dir=_test_dir)
    except (name_resolve.NameResolveError, timeout):
        pytest.xfail("Simbad is down")

    base, ext = path.splitext(full_path)
    if new_file_ext is None:
        new_ext = 'new'  # this is the default value...
    else:
        new_ext = new_file_ext
    new_path = base + new_ext + ext
    f = fits.open(new_path, do_not_scale_image_data=True)
    h = f[0].header
    m101_ra_dec_correct = FK5('14h03m12.58s +54d20m55.50s')
    header_m101 = FK5(ra=h['ra'], dec=h['dec'],
                      unit=(u.hour, u.degree))

    assert_almost_equal(m101_ra_dec_correct.ra.hour,
                        header_m101.ra.hour, decimal=6)
    assert_almost_equal(m101_ra_dec_correct.dec.degree,
                        header_m101.dec.degree, decimal=6)


def test_add_ra_dec_from_object_name_edge_cases(caplog):
    # add a 'dec' keyword to every light file so that none need RA/Dec
    ic = ImageFileCollection(_test_dir, keywords=['imagetyp'])
    for h in ic.headers(imagetyp='light', overwrite=True):
        h['dec'] = '+17:42:00'
        h['ra'] = '03:45:06'
        h['object'] = 'm101'
    # does this succeed without errors?
    ph.add_ra_dec_from_object_name(_test_dir)

    # add name that will fail as object of one image
    image_path = path.join(_test_dir, _test_image_name)
    f = fits.open(image_path)
    h = f[0].header
    del h['RA']
    del h['dec']
    h['object'] = 'i am a fake object'
    f.writeto(image_path, clobber=True)
    ph.add_ra_dec_from_object_name(_test_dir)
    warns = get_patch_header_logs(caplog, level=logging.WARN)
    assert 'Unable to lookup' in warns


def get_patch_header_logs(log, level=logging.WARN):
    patch_header_warnings = []
    for record in log.records():
        if (('patchers' in record.name) and (record.levelno ==
                                             level)):

            patch_header_warnings.append(record.message)

    patch_headers_message_text = '\n'.join(patch_header_warnings)
    return patch_headers_message_text


def test_times_apparent_pos_added():
    # the correct value below is from the USNO JD calculator using the UT
    # of the start of the observation in the file uint16.fit, which is
    # 2012-06-05T04:17:00
    JD_correct = 2456083.678472222
    # The "correct" values below are from converting the RA/Dec in uint16 to
    # decimal hours/degrees
    RA_correct = 14.054166667  # hours
    RA_2012_5 = 14.061475336  # hours, from astropy
    Dec_correct = 54.351111111  # degrees
    Dec_2012_5 = 54.29181012  # degrees, from astropy
    # Got the "correct" LST from astropy
    LST = 14.78635133  # hours

    HA_correct = LST - RA_correct

    ph.patch_headers(_test_dir, new_file_ext='', overwrite=True)
    header = fits.getheader(path.join(_test_dir, _test_image_name))

    # check Julian Date
    assert_almost_equal(header['JD-OBS'], JD_correct)

    # calculate then check LST
    h, m, s = header['LST'].split(':')
    header_lst = int(h) + int(m)/60. + float(s)/3600.

    assert_almost_equal(header_lst, LST, decimal=7)

    # calculate then check HA
    h, m, s = header['HA'].split(':')
    header_HA = int(h) + int(m)/60. + float(s)/3600.

    assert_almost_equal(HA_correct, header_HA, decimal=7)

    print(header['MJD-OBS'])
    # calculate, then check, altitude
    f = FederSite()
    latitude = f.latitude.radian
    Dec_a = Angle(Dec_2012_5, unit=u.degree)
    HA_a = Angle(LST - RA_2012_5, unit=u.hour)
    sin_alt = (np.sin(latitude) * np.sin(Dec_a.radian) +
               (np.cos(latitude) * np.cos(Dec_a.radian) *
                np.cos(HA_a.radian)))
    alt = Angle(np.arcsin(sin_alt), unit=u.radian)
    header_alt = header['alt-obj']

    # ONLY CHECKING TWO DECIMAL PLACES SEEMS AWFUL!!!
    assert_almost_equal(alt.degree, header_alt, decimal=2)

    # calculate, then check, airmass
    zenith_angle = Angle(90 - alt.degree, unit=u.degree)
    airmass_correct = 1/np.cos(zenith_angle.radian)
    assert_almost_equal(airmass_correct, header['airmass'], decimal=3)


def test_add_object_pos_airmass_raises_error_when_it_should():
    feder = Feder()
    header = fits.Header()
    # make sure the JD_OBS really has the desired value for this test...
    assert feder.JD_OBS.value is None
    # have not set JD_OBS, do we raise the appropriate error?
    with pytest.raises(ValueError):
        ph.add_object_pos_airmass(header)
    # No significance to the value below, but need to set it to something...
    feder.JD_OBS.value = 2456490
    with pytest.raises(ValueError):
        ph.add_object_pos_airmass(header)


def test_purge_handles_all_software():
    ic = ImageFileCollection(_test_dir, keywords=['imagetyp'])
    for h in ic.headers():
        ph.purge_bad_keywords(h)
        assert 'purged' in h


def test_purge_bad_keywords_logic_for_conditionals(caplog):
    ic = ImageFileCollection(_test_dir, keywords=['imagetyp'])
    headers = [h for h in ic.headers()]
    a_header = headers[0]
    # should be no warnings the first time...
    ph.purge_bad_keywords(a_header)
    purge_warnings = get_patch_header_logs(caplog)
    assert not purge_warnings
    # re-purging should generate warnings...
    ph.purge_bad_keywords(a_header)
    purge_warnings = get_patch_header_logs(caplog)
    assert 'force' in purge_warnings
    assert 'removing' in purge_warnings
    # grab a clean header
    # want to get to a header with more than one bad keyword so that a
    # history is generated...
    for a_header in headers[1:]:
        software = ph.get_software_name(a_header)
        if len(software.bad_keywords) <= 1:
            continue
        else:
            break
    # delete one of the purge keywords for this software, which should ensure
    # that their is no history added to the header that contains the name of
    # this keyword
    key_to_delete = software.bad_keywords[0]
    print(software.bad_keywords)
    del a_header[key_to_delete]
    print(a_header)
    ph.purge_bad_keywords(a_header, history=True)
    print(a_header)
    assert all(key_to_delete.lower() not in h.lower()
               for h in a_header['HISTORY'])


def test_unit_is_added():
    # patch in _test_dir, overwriting existing files
    ph.patch_headers(_test_dir, overwrite=True, new_file_ext='')
    ic = ImageFileCollection(_test_dir, keywords='*')
    feder = Feder()
    print(_test_dir)
    for h, f in ic.headers(return_fname=True):
        instrument = feder.instruments[h['instrume']]
        if instrument.image_unit is not None:
            print(str(instrument.image_unit), f)
            # If the instrument has a unit, the header should too.
            assert h['BUNIT'] == str(instrument.image_unit)


def test_lst_in_future():
    header = fits.Header()
    future = Time.now() + 1 * u.year
    header['date-obs'] = future.isot
    ph.add_time_info(header)


@pytest.fixture(params=['object', 'OBJECT', 'Object'])
def object_file_ra_change_col_case(request):
    object_file_with_ra_dec(_test_dir, object_col_name=request.param)


@pytest.fixture
def object_file_no_ra(request):
    try:
        object_col_name = request.param
    except AttributeError:
        object_col_name = 'object'
    to_write = ('# comment 1\n# comment 2\n' + object_col_name +
                '\ney uma\nm101\n')
    object_file = open(path.join(_test_dir, _default_object_file_name), 'wt')
    object_file.write(to_write)
    object_file.close()


def object_file_with_ra_dec(dir, object_col_name='object',
                            input_objects=None):
    if input_objects is None:
        objs = ["ey uma, 09:02:20.76, +49:49:09.3",
                "m101,14:03:12.58,+54:20:55.50"
                ]
    else:
        objs = list(input_objects)

    to_write = ('# comment 1\n# comment 2\n' + object_col_name +
                ', RA, Dec\n' + '\n'.join(objs))
    object_file = open(path.join(dir, _default_object_file_name), 'wt')
    object_file.write(to_write)
    object_file.close()


def setup_module(module):
    global simbad_down

    try:
        foo = SkyCoord.from_name("m101")
    except (name_resolve.NameResolveError, timeout):
        simbad_down = True


def setup_function(function):
    global _test_dir
    global _test_image_name

    _test_dir = path.join(mkdtemp(), 'data')
    data_source = get_data_dir()
    print(data_source)
    copytree(data_source, _test_dir)
    object_file_with_ra_dec(_test_dir)


def teardown_function(function):
    global _test_dir
    rmtree(_test_dir)



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

import os
from contextlib import contextmanager

import astropy.io.fits as fits
from astropy.table import Table, Column
from astropy.extern import six

import pytest
import py
import numpy as np

from ...header_processing.patchers import IRAF_image_type
from .. import run_patch as run_patch

from .. import run_triage

from .. import run_astrometry
from .. import run_standard_header_process
from .. import sort_files
from ...image_collection import ImageFileCollection
from ..script_helpers import handle_destination_dir_logging_check
from .. import quick_add_keys_to_file
from ...tests.data import get_data_dir

_default_object_file_name = 'obsinfo.txt'
OBJECT_LIST_URL = 'https://raw.github.com/mwcraig/feder-object-list/master/feder_object_list.csv'


def set_mtimes(files, offset=10):
    """
    Set mtime of files to a time in the past

    Parameters
    ----------

    files : list of py.path.local objects
        list of files for which mtime should be set
    """
    default_mtime = files[0].mtime() - offset
    for fil in files:
        fil.setmtime(default_mtime)


def mtimes(files):
    """
    Find mtimes of files

    Parameters
    ----------

    files : list of py.path.local objects
        list of files
    """
    return [fil.mtime() for fil in files]


@pytest.fixture
def triage_dict():
    default_names = run_triage.DefaultFileNames()
    names_dict = default_names.as_dict()
    return names_dict


@pytest.fixture
def clean_data(tmpdir, request):
    test_dir = tmpdir
    test_data_dir = py.path.local(get_data_dir())
    test_data_dir.copy(test_dir)
    objs = ["ey uma, 09:02:20.76, +49:49:09.3",
            "m101,14:03:12.58,+54:20:55.50"
            ]
    to_write = '# comment 1\n# comment 2\nobject, RA, Dec\n' + '\n'.join(objs)
    object_path = test_dir.join(_default_object_file_name)
    print(object_path)
    object_file = object_path.open(mode='wt')
    print(object_file)
    object_file.write(to_write)
    object_file.close()

    def cleanup():
        test_dir.remove()
    request.addfinalizer(cleanup)
    return test_dir


@pytest.mark.usefixtures('clean_data')
class TestScript(object):
    @pytest.fixture(autouse=True)
    def set_test_dir(self, clean_data):
        self.test_dir = clean_data

    @pytest.fixture
    def default_keywords(self):
        return run_triage.DEFAULT_KEYS

    def test_run_patch_does_not_overwite_fits_if_dest_set(self, recwarn):
        fits_files = [fil for fil in self.test_dir.visit(fil=str('*.fit'),
                                                         sort=False)]
        set_mtimes(fits_files)
        original_mtimes = mtimes(fits_files)
        print(original_mtimes)
        destination = self.test_dir.make_numbered_dir()
        arglist = ['--destination-dir', destination.strpath,
                   self.test_dir.strpath
                   ]
        run_patch.main(arglist)
        new_mtimes = mtimes(fits_files)
        print(new_mtimes)
        print(destination)
        print(self.test_dir)
        for original, new in zip(original_mtimes, new_mtimes):
            assert (original == new)
        # assertion below is to catch whether add_object_info raised a warning
        # if it did, it means object names were not added and the test wasn't
        # complete
        #
        # originally tested for this by looking at number of warnings
        # emitted, but changing astropy versions made it emit warnings
        # instead, do the stupid thing below...check for text of the message.
        #
        for wrn in recwarn.list:
            # Apparently message is not actually always a string, so the
            # assert below can fail for reasons unrelated to the test.
            try:
                assert 'No object list' not in wrn.message
            except TypeError:
                pass
        destination.remove()

    def test_run_patch_with_object_list_url(self, simbad_down):
        if simbad_down:
            pytest.xfail("simbad is down")
        arglist = ['-o', OBJECT_LIST_URL, self.test_dir.strpath]
        # Remove default object file just to make sure object information can
        # come only from the remote list.
        self.test_dir.join(_default_object_file_name).remove()
        run_patch.main(arglist)
        h = fits.getheader(self.test_dir.join('uint16.fit').strpath)
        assert h['object'] == 'm101'

    def test_run_patch_overscan_only(self, simbad_down):
        if simbad_down:
            pytest.xfail("simbad is down")
        arglist = ['-o', OBJECT_LIST_URL, '--overscan-only',
                   self.test_dir.strpath]
        # Remove default object file just to make sure object information can
        # come only from the remote list.
        self.test_dir.join(_default_object_file_name).remove()
        run_patch.main(arglist)
        h = fits.getheader(self.test_dir.join('uint16.fit').strpath)
        print(h)
        assert 'object' not in h

    def test_run_triage_no_output_generated(self, default_keywords):
        list_before = self.test_dir.listdir(sort=True)
        run_triage.triage_directories([self.test_dir.strpath],
                                      keywords=default_keywords,
                                      object_file_name=None,
                                      pointing_file_name=None,
                                      filter_file_name=None,
                                      output_table=None)
        list_after = self.test_dir.listdir(sort=True)
        assert(list_before == list_after)

    def _verify_triage_files_created(self, directory, triage_dict):
        for option_name, file_name in six.iteritems(triage_dict):
            print(option_name, file_name, directory.join(file_name).check())
            assert(directory.join(file_name).check())

    def test_triage_output_file_by_keyword(self, triage_dict,
                                           default_keywords):
        run_triage.triage_directories([self.test_dir.strpath],
                                      keywords=default_keywords,
                                      **triage_dict)
        self._verify_triage_files_created(self.test_dir, triage_dict)

    def test_triage_destination_directory(self, triage_dict,
                                          default_keywords):
        destination = self.test_dir.make_numbered_dir()
        list_before = self.test_dir.listdir(sort=True)
        run_triage.triage_directories([self.test_dir.strpath],
                                      keywords=default_keywords,
                                      destination=destination.strpath,
                                      **triage_dict)
        list_after = self.test_dir.listdir(sort=True)
        assert(list_before == list_after)
        self._verify_triage_files_created(destination, triage_dict)

    @pytest.mark.parametrize('extra_keys',
                             [['key1'],
                              ['key1', 'key2']])
    def test_run_triage_default_keys_and_extra(self, default_keywords,
                                               triage_dict, extra_keys):
        table_file = 'tbl.txt'
        arglist = []
        print(extra_keys)
        for key in extra_keys:
            arglist.extend(['--key', key])

        gather_keys = default_keywords[:]

        gather_keys.extend(extra_keys)
        arglist.extend(['--table-name', table_file,
                       self.test_dir.strpath])
        run_triage.main(arglist=arglist)
        print(arglist)
        print(gather_keys)
        table = Table.read(self.test_dir.join(table_file).strpath,
                           format='ascii')
        for key in gather_keys:
            assert (key in table.colnames)

    def test_run_triage_on_set_with_no_light_files(self):
        ic = ImageFileCollection(self.test_dir.strpath, keywords=['imagetyp'])
        for header in ic.headers(imagetyp='light', overwrite=True):
            header['imagetyp'] = 'BIAS'
        arglist = [self.test_dir.strpath]
        run_triage.main(arglist)
        assert 1

    def test_run_triage_on_set_with_maximdl_imagetype_fails(self):
        hdu = fits.PrimaryHDU()
        hdu.header['imagetyp'] = 'Bias Frame'
        hdu.header['exptime'] = 0.0
        hdu.data = np.random.random([100, 100])
        hdu.writeto(self.test_dir.join('maxim.fits').strpath)
        with pytest.raises(ValueError):
            run_triage.triage_fits_files(self.test_dir.strpath)

    def test_run_triage_contains_columns_with_extended_location_info(self):
        result = run_triage.triage_fits_files(self.test_dir.strpath)
        location_keys = ['Source path', 'Source directory']
        for key in location_keys:
            assert key in result['files'].colnames

    def test_run_triage_correctly_sets_extended_location_info(self):
        result = run_triage.triage_fits_files(self.test_dir.strpath)
        table = result['files']
        assert table['Source path'][0] == self.test_dir.strpath
        assert (table['Source directory'][0] ==
                os.path.basename(self.test_dir.strpath))

    def test_run_triage_with_only_list_default_keys(self):
        assert run_triage.main(arglist=['-l']) == run_triage.DEFAULT_KEYS
        assert (run_triage.main(arglist=['-l', '-k should_not_be_added']) ==
                run_triage.DEFAULT_KEYS)

    def test_run_triage_raises_error_if_no_dir_supplied(self):
        with pytest.raises(SystemExit):
            # dummy argument to ensure coverage knows the test is happening
            run_triage.main(arglist=['-k cow'])

    def test_run_triage_with_no_keywords_makes_right_keywords(self):
        result = run_triage.triage_fits_files(self.test_dir.strpath)
        expected_keys = ['imagetyp', 'object', 'filter']
        for key in expected_keys:
            assert key in result['files'].colnames

    def test_triage_grabbing_all_keywords_gets_them_all(self):
        tbl_name = 'tbl.txt'
        run_triage.main(arglist=['-a', '-t', tbl_name, self.test_dir.strpath])
        rt_table = Table.read(self.test_dir.join(tbl_name).strpath,
                              format='ascii.csv')
        lcase_columns = [c.lower() for c in rt_table.colnames]
        print(lcase_columns)
        ic = ImageFileCollection(self.test_dir.strpath,
                                 keywords='*')
        for h in ic.headers():
            for k in h:
                if k:
                    assert k.lower() in lcase_columns

    def test_run_triage_writer_makes_correct_column_names(self, tmpdir):
        dump_file = 'dumb.txt'
        run_triage.write_list(tmpdir.strpath, dump_file, range(10))
        tab = Table.read(tmpdir.join(dump_file).strpath, format='ascii')
        assert 'File' in tab.colnames

        custom_name = 'Bob'
        run_triage.write_list(tmpdir.strpath, dump_file, range(10),
                              column_name=custom_name)
        tab = Table.read(tmpdir.join(dump_file).strpath, format='ascii')
        assert custom_name in tab.colnames

    def test_triage_case_inseneistive_column_name_matching(self):
        columns = ['one', 'Two', 'THREE']
        assert (run_triage.get_column_name_case_insensitive('one', columns)
                == 'one')
        assert (run_triage.get_column_name_case_insensitive('NOT', columns)
                == '')
        assert (run_triage.get_column_name_case_insensitive('two', columns)
                == 'Two')

    def test_run_astrometry_with_dest_does_not_modify_source(self):

        destination = self.test_dir.make_numbered_dir()
        list_before = self.test_dir.listdir(sort=True)
        arglist = ['--destination-dir', destination.strpath,
                   self.test_dir.strpath
                   ]
        run_astrometry.main(arglist)
        list_after = self.test_dir.listdir(sort=True)
        # nothing should change in the source directory
        assert(list_before == list_after)
        # for each light file in the destination directory we should have a
        # file with the same basename but an extension of blind
        ic = ImageFileCollection(destination.strpath,
                                 keywords=['IMAGETYP'])
        for image in ic.files_filtered(imagetyp='LIGHT'):
            image_path = destination.join(image)
            print(image_path.purebasename)
            blind_path = destination.join(image_path.purebasename + '.blind')
            print(blind_path.strpath)
            assert (blind_path.check())

    @pytest.mark.parametrize('file_column',
                             ['file',
                              'FiLe',
                              'badd'])
    @pytest.mark.parametrize('file_arg',
                             ['--file-list',
                              ''])
    @pytest.mark.parametrize('keyword_arg',
                             ['--key-file',
                              '--key-value'])
    def test_quick_add_keys_records_history(self, keyword_arg,
                                            file_arg, file_column):
        ic = ImageFileCollection(self.test_dir.strpath,
                                 keywords=['imagetyp'])
        ic.summary_info.keep_columns('file')

        file_list = os.path.join(ic.location, 'files.txt')
        keyword_list = os.path.join(ic.location, 'keys.txt')

        full_paths = [os.path.join(self.test_dir.strpath, fil) for
                      fil in ic.summary_info['file']]
        print('fill paths: %s' % ' '.join(full_paths))
        ic.summary_info['file'][:] = full_paths
        ic.summary_info.remove_column('file')
        ic.summary_info.add_column(Column(data=full_paths, name=file_column))
        ic.summary_info.write(file_list, format='ascii')
        if file_column != 'file':
            ic.summary_info.rename_column(file_column, 'file')
        dumb_keyword = 'munkeez'.upper()
        dumb_value = 'bananaz'
        keywords = Column(data=[dumb_keyword], name='Keyword')
        vals = Column(data=[dumb_value], name='value')
        keyword_table = Table()
        keyword_table.add_columns([keywords, vals])
        keyword_table.write(keyword_list, format='ascii')
        args_for = {}
        args_for['--key-file'] = [keyword_list]
        args_for['--key-value'] = [dumb_keyword, dumb_value]
        args_for['--file-list'] = [file_list]
        args_for[''] = full_paths
        argslist = [keyword_arg]
        argslist.extend(args_for[keyword_arg])
        if file_arg:
            argslist.append(file_arg)
        argslist.extend(args_for[file_arg])
        if file_column.lower() != 'file' and file_arg:
            with pytest.raises(ValueError):
                quick_add_keys_to_file.main(argslist)
            return
        else:
            quick_add_keys_to_file.main(argslist)

#        add_keys(file_list=file_list, key_file=keyword_list)
        for header in ic.headers():
            assert (header[dumb_keyword] == dumb_value)
            history_string = ' '.join(header['history'])
            assert (dumb_keyword in history_string)
            assert (dumb_value in history_string)

    def test_quick_add_keys_raises_error_if_no_files(self):
        with pytest.raises(SystemExit):
            quick_add_keys_to_file.main(['--key-value', 'key', 'value'])


@pytest.fixture(params=['run_patch', 'run_triage', 'run_astrometry'])
def a_parser(request):
    from ..run_patch import construct_parser as ph_parser
    from ..run_astrometry import construct_parser as as_parser
    from ..run_triage import construct_parser as tr_parser
    parsers = {'run_patch': ph_parser,
               'run_astrometry': as_parser,
               'run_triage': tr_parser
               }
    the_parser = parsers[request.param]
    return the_parser()

## snippet below is from http://stackoverflow.com/a/13847807


@contextmanager
def pushd(newDir):
    previousDir = os.getcwd()
    os.chdir(newDir)
    yield
    os.chdir(previousDir)


class TestScriptHelper(object):
    """Test functions in script_helpers"""

    @pytest.mark.parametrize("argstring,run_in,expected", [
        (['--no-log-destination', '--destination-dir', '.', '.'], '.', 'exception'),
        (['--no-log-destination', os.getcwd()], '.', 'exception'),
        (['--no-log-destination', '--destination-dir', '/tmp', os.getcwd()], '.', True),
        (['--destination-dir', '/tmp', '.'], '.', False),
        (['--destination-dir', os.getcwd(), os.getcwd()], '..', False)])
    def test_handle_destination_dir_logging_check(self, argstring, run_in,
                                                  expected, a_parser):
        with pushd(run_in):
            args = a_parser.parse_args(argstring)
            print(argstring, expected)
            print(type(expected))
            if expected == 'exception':
                with pytest.raises(RuntimeError):
                    handle_destination_dir_logging_check(args)
            else:
                assert (handle_destination_dir_logging_check(args) == expected)


@pytest.mark.usefixtures('clean_data')
class TestRunStandardHeaderProcess(object):
    @pytest.fixture(autouse=True)
    def set_test_dir(self, clean_data):
        self.test_dir = clean_data

    @pytest.fixture()
    def scratch_destination(self):
        dest = self.test_dir.mkdtemp()
        return dest

    def test_source_not_modified_if_dest_set(self, scratch_destination):
        arglist = ['--dest-root', scratch_destination.strpath]
        arglist += [self.test_dir.strpath]
        fits_files = [fil for fil in self.test_dir.visit(fil=str('*.fit'),
                                                         sort=True)]
        set_mtimes(fits_files)
        original_mtimes = mtimes(fits_files)
        run_standard_header_process.main(arglist)
        new_mtimes = mtimes(fits_files)
        for original, new in zip(original_mtimes, new_mtimes):
            assert(original == new)


class TestSortFiles(object):
    """docstring for TestSortFiles"""
    @pytest.fixture(autouse=True)
    def set_test_dir(self, tmpdir):
        self.test_dir = tmpdir

    def write_names(self, hdu, names):
        for name in names:
            hdu.writeto(self.test_dir.join(name).strpath)

    def make_filter_exp(self, hdu, filter_dict, extra_name=None):
        add_to_name = extra_name or ''
        hdr = hdu.header
        make_base = lambda hdr, band: \
            hdr['imagetyp'] + add_to_name + band + str(hdr['exptime'])
        for band, num_or_exp_dict in six.iteritems(filter_dict):
            hdr['filter'] = band
            if isinstance(num_or_exp_dict, dict):
                for exp, number in six.iteritems(num_or_exp_dict):
                    hdr['exptime'] = exp
                    base = make_base(hdr, band)
                    print(base)
                    self.write_names(hdu, [base + str(i) + '.fit' for i in range(number)])
            else:
                hdr['exptime'] = 17.0
                base = make_base(hdr, band)
                print(base)
                print(hdr['imagetyp'])
                self.write_names(hdu, [base + str(i) + '.fit' for i in range(num_or_exp_dict)])

    @pytest.fixture
    def set_test_files(self):
        """
        A set of file is created from the dictionary below; the number
        of files of each type is set by the 'number' key.
        """
        bias = 'BIAS'
        dark = 'DARK'
        flat = 'FLAT'
        light = 'LIGHT'
        files = {bias: 5,
                 dark: {20.0: 5,
                        30.0: 5},
                 flat: {'R': {17.0: 3},
                        'V': {15.0: 2,
                              25.0: 2},
                        'I': {17.0: 5}},
                 light: {'m81': {'R': {17.0: 3},
                                 'V': {17.0: 2}},
                         'm101': {'I': {17.0: 5}},
                         None: {'I': {17.0: 2},
                                'R': {17.0: 3},
                                'V': {17.0: 4}}}
                 }
        data = np.random.random([100, 100])
        make_hdu = lambda im_type: \
            fits.PrimaryHDU(data,
                            fits.Header.fromkeys(['imagetyp'], value=im_type))
        make_file_names = lambda base, number: \
            [base + str(i) + '.fit' for i in range(number)]
        bias_names = make_file_names(bias, files[bias])
        bias_hdu = make_hdu(bias)
        self.write_names(bias_hdu, bias_names)
        for exp, number in six.iteritems(files[dark]):
            dark_hdu = make_hdu(dark)
            dark_hdu.header['exptime'] = exp
            dark_names = make_file_names(dark + str(exp), number)
            self.write_names(dark_hdu, dark_names)
        flat_hdu = make_hdu(flat)
        self.make_filter_exp(flat_hdu, files[flat])
        light_hdu = make_hdu(light)
        for obj, filter_dict in six.iteritems(files[light]):
            if not obj:
                del light_hdu.header['object']
            else:
                light_hdu.header['object'] = obj
            obj_name = obj or None
            self.make_filter_exp(light_hdu, filter_dict, extra_name=obj_name)

        return files

    def test_sort_files_can_be_called(self):
        sort_files.main([self.test_dir.strpath])

    def test_sort_called_with_no_directory_raises_error(self):
        with pytest.raises(SystemExit):
            sort_files.main(arglist=['-v', '-n'])

    def _walk_dict_string_keys(self, d, parent=None):
        if parent is None:
            parent = []
        for key, val in six.iteritems(d):
            new_parent = list(parent) + [str(key)]
            if isinstance(val, dict):
                for d in self._walk_dict_string_keys(val, parent=new_parent):
                    yield d
            else:
                yield new_parent, val

    @pytest.mark.parametrize('dest,move',
                             [('tmp', False),
                              (None, False),
                              ('tmp', True),
                              (None, True)])
    def test_sort_called_with_destination(self, set_test_files, dest, move):
        print(set_test_files)
        if dest:
            dest_path = self.test_dir.mkdtemp().strpath
            args = ['-d', dest_path, self.test_dir.strpath]
        else:
            dest_path = self.test_dir.strpath
            args = [self.test_dir.strpath]
        if move:
            args.insert(0, '--move')
        files_before_sort = os.listdir(self.test_dir.strpath)
        sort_files.main(arglist=args)
        files_after_sort = os.listdir(self.test_dir.strpath)
        if dest and not move:
            assert files_before_sort == files_after_sort
        if move:
            for f in files_after_sort:
                print(f)
                assert os.path.isdir(os.path.join(self.test_dir.strpath, f))
        expected_unsorted = 0
        for key in set_test_files['LIGHT'][None]:
            expected_unsorted += set_test_files['LIGHT'][None][key][17.0]
        for parents, num_files in self._walk_dict_string_keys(set_test_files):
            print(parents)
            if 'None' in parents:
                path = os.path.join(dest_path, parents[0],
                                    sort_files.UNSORTED_DIR)
                assert len(os.listdir(path)) == expected_unsorted
            else:
                path = os.path.join(dest_path, *parents)
                assert len(os.listdir(path)) == num_files

    def test_sort_creates_destination_if_needed(self, set_test_files):
        dest = self.test_dir.mkdtemp()
        dest = dest.join('crazy_dir')
        args = ['-d', dest.strpath, self.test_dir.strpath]
        sort_files.main(arglist=args)
        assert os.path.isdir(dest.strpath)

    def test_sort_handles_cannot_form_tree(self, set_test_files):
        # the object keyword is stripped from all headers, which means
        # you cannot for a tree for the light files. As a result, all
        # light files should end up in "unsorted" and no error should
        # be raised.
        images = ImageFileCollection(self.test_dir.strpath,
                                     keywords=['imagetyp', 'object'])
        n_light = 0
        for header in images.headers(overwrite=True, imagetyp='LIGHT'):
            del header['object']
            n_light += 1
        dest = self.test_dir.mkdtemp()
        sort_files.main(['-d', dest.strpath, self.test_dir.strpath])
        unsorted_path = os.path.join(dest.strpath,
                                     'LIGHT', sort_files.UNSORTED_DIR)
        assert len(os.listdir(unsorted_path)) == n_light


def test_triage_via_triage_fits_files(triage_setup):
    file_info = run_triage.triage_fits_files(triage_setup.test_dir)
    print("number of files should be %i" % triage_setup.n_test['files'])
    print(file_info['files']['file'])
    assert len(file_info['files']['file']) == triage_setup.n_test['files']
    assert len(file_info['needs_pointing']) == \
        triage_setup.n_test['need_pointing']
    assert len(file_info['needs_object_name']) == \
        triage_setup.n_test['need_object']
    assert len(file_info['needs_filter']) == triage_setup.n_test['need_filter']
    assert len(file_info['needs_astrometry']) == triage_setup.n_test['light']
    bias_check = np.where(file_info['files']['imagetyp'] ==
                          IRAF_image_type('bias'))
    assert (len(bias_check[0]) == 2)


def test_triage_via_run_triage(triage_setup, triage_dict):
    run_triage.main(['-a', triage_setup.test_dir])
    n_test_names = ['need_pointing', 'need_object', 'need_filter']
    file_name_keys = ['pointing_file_name', 'object_file_name',
                      'filter_file_name']
    for n_name, fname in zip(n_test_names, file_name_keys):
        file_path = os.path.join(triage_setup.test_dir,
                                 triage_dict[fname])
        dump = Table.read(file_path, format='ascii')
        assert len(dump) == triage_setup.n_test[n_name]



from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

import os
from shutil import rmtree
from tempfile import mkdtemp
from glob import iglob, glob
import logging
import stat

import astropy.io.fits as fits
import numpy as np
import pytest

from .. import image_collection as tff
from ..header_processing.patchers import IRAF_image_type

_filters = []
_original_dir = ''


def test_fits_summary(triage_setup):
    keywords = ['imagetyp', 'filter']
    image_collection = tff.ImageFileCollection(triage_setup.test_dir,
                                               keywords=keywords)
    summary = image_collection._fits_summary(header_keywords=keywords)
    print(summary['file'])
    print(summary.keys())
    assert len(summary['file']) == triage_setup.n_test['files']
    for keyword in keywords:
        assert len(summary[keyword]) == triage_setup.n_test['files']
    # explicit conversion to array is needed to avoid astropy Table bug in
    # 0.2.4
    print(np.array(summary['file'] == 'no_filter_no_object_bias.fit'))
    no_filter_no_object_row = np.array(summary['file'] ==
                                       'no_filter_no_object_bias.fit')
    # there should be no filter keyword in the bias file
    assert (summary['filter'][no_filter_no_object_row].mask)


# this should work, but doesn't:
# @pytest.mark.usefixtures("triage_setup")
class TestImageFileCollection(object):
    def test_filter_files(self, triage_setup):
        img_collection = tff.ImageFileCollection(
            location=triage_setup.test_dir, keywords=['imagetyp', 'filter'])
        print(img_collection.files_filtered(imagetyp='bias'))
        print(triage_setup.n_test)
        assert len(img_collection.files_filtered(
            imagetyp='bias')) == triage_setup.n_test['bias']
        assert len(img_collection.files) == triage_setup.n_test['files']
        assert ('filter' in img_collection.keywords)
        assert ('flying monkeys' not in img_collection.keywords)
        assert len(img_collection.values('imagetyp', unique=True)) == 2

    def test_summary_is_summary_info(self, triage_setup):
        img_collection = tff.ImageFileCollection(
            location=triage_setup.test_dir, keywords=['imagetyp', 'filter'])
        assert img_collection.summary is img_collection.summary_info

    def test_files_with_compressed(self, triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir)
        assert len(collection._fits_files_in_directory(
            compressed=True)) == triage_setup.n_test['files']

    def test_files_with_no_compressed(self, triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir)
        n_files_found = len(
            collection._fits_files_in_directory(compressed=False))
        n_uncompressed = (triage_setup.n_test['files'] -
                          triage_setup.n_test['compressed'])
        assert n_files_found == n_uncompressed

    def test_generator_full_path(self, triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir,
                                             keywords=['imagetyp'])
        for path, file_name in zip(collection._paths(), collection.files):
            assert path == os.path.join(triage_setup.test_dir, file_name)

    def test_hdus(self, triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir,
                                             keywords=['imagetyp'])
        n_hdus = 0
        for hdu in collection.hdus():
            assert isinstance(hdu, fits.PrimaryHDU)
            data = hdu.data  # must access the data to force scaling
            with pytest.raises(KeyError):
                hdu.header['bzero']
            n_hdus += 1
        assert n_hdus == triage_setup.n_test['files']

    def test_hdus_masking(self, triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir,
                                             keywords=['imagetyp', 'exposure'])
        old_data = np.array(collection.summary_info)
        for hdu in collection.hdus(imagetyp='bias'):
            pass
        new_data = np.array(collection.summary_info)
        assert (new_data == old_data).all()

    def test_headers(self, triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir,
                                             keywords=['imagetyp'])
        n_headers = 0
        for header in collection.headers():
            assert isinstance(header, fits.Header)
            assert ('bzero' in header)
            n_headers += 1
        assert n_headers == triage_setup.n_test['files']

    def test_headers_save_location(self, triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir,
                                             keywords=['imagetyp'])
        destination = mkdtemp()
        for header in collection.headers(save_location=destination):
            pass
        new_collection = \
            tff.ImageFileCollection(location=destination,
                                    keywords=['imagetyp'])
        basenames = lambda paths: set(
            [os.path.basename(file) for file in paths])

        assert (len(basenames(collection._paths()) -
                    basenames(new_collection._paths())) == 0)
        rmtree(destination)

    def test_headers_with_filter(self, triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir,
                                             keywords=['imagetyp'])
        cnt = 0
        for header in collection.headers(imagetyp='light'):
            assert header['imagetyp'].lower() == 'light'
            cnt += 1
        assert cnt == triage_setup.n_test['light']

    def test_headers_with_multiple_filters(self, triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir,
                                             keywords=['imagetyp'])
        cnt = 0
        for header in collection.headers(imagetyp='light',
                                         filter='R'):
            assert header['imagetyp'].lower() == 'light'
            assert header['filter'].lower() == 'r'
            cnt += 1
        assert cnt == (triage_setup.n_test['light'] -
                       triage_setup.n_test['need_filter'])

    def test_headers_with_filter_wildcard(self, triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir,
                                             keywords=['imagetyp'])
        cnt = 0
        for header in collection.headers(imagetyp='*'):
            cnt += 1
        assert cnt == triage_setup.n_test['files']

    def test_headers_with_filter_missing_keyword(self, triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir,
                                             keywords=['imagetyp'])
        for header in collection.headers(imagetyp='light',
                                         object=''):
            assert header['imagetyp'].lower() == 'light'
            with pytest.raises(KeyError):
                header['object']

    def test_generator_headers_save_with_name(self, triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir,
                                             keywords=['imagetyp'])
        for header in collection.headers(save_with_name='_new'):
            assert isinstance(header, fits.Header)
        new_collection = \
            tff.ImageFileCollection(location=triage_setup.test_dir,
                                    keywords=['imagetyp'])
        assert (len(new_collection._paths()) ==
                2 * (triage_setup.n_test['files']) -
                triage_setup.n_test['compressed'])
        print(glob(triage_setup.test_dir + '/*_new*'))
        [os.remove(fil) for fil in iglob(triage_setup.test_dir + '/*_new*')]
        print(glob(triage_setup.test_dir + '/*_new*'))

    def test_generator_data(self, triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir,
                                             keywords=['imagetyp'])
        for img in collection.data():
            assert isinstance(img, np.ndarray)

    def test_consecutive_fiilters(self, triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir,
                                             keywords=['imagetyp',
                                                       'filter',
                                                       'object'])
        no_files_match = collection.files_filtered(object='fdsafs')
        assert(len(no_files_match) == 0)
        some_files_should_match = collection.files_filtered(object=None,
                                                            imagetyp='light')
        print(some_files_should_match)
        assert(len(some_files_should_match) ==
               triage_setup.n_test['need_object'])

    def test_filter_does_not_not_permanently_change_file_mask(self,
                                                              triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir,
                                             keywords=['imagetyp'])
        # ensure all files are originally unmasked
        assert(not collection.summary_info['file'].mask.any())
        # generate list that will match NO files
        collection.files_filtered(imagetyp='foisajfoisaj')
        # if the code works, this should have no permanent effect
        assert(not collection.summary_info['file'].mask.any())

    @pytest.mark.parametrize("new_keywords,collection_keys", [
                            (['imagetyp', 'object'], ['imagetyp', 'filter']),
                            (['imagetyp'], ['imagetyp', 'filter'])])
    def test_keyword_setting(self, new_keywords, collection_keys,
                             triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir,
                                             keywords=collection_keys)
        tbl_orig = collection.summary_info
        collection.keywords = new_keywords
        tbl_new = collection.summary_info

        if set(new_keywords).issubset(collection_keys):
            # should just delete columns without rebuilding table
            assert(tbl_orig is tbl_new)
        else:
            # we need new keywords so must rebuild
            assert(tbl_orig is not tbl_new)

        for key in new_keywords:
            assert(key in tbl_new.keys())
        assert (tbl_orig['file'] == tbl_new['file']).all()
        assert (tbl_orig['imagetyp'] == tbl_new['imagetyp']).all()
        assert 'filter' not in tbl_new.keys()
        assert 'object' not in tbl_orig.keys()

    def test_keyword_setting_to_empty_list(self, triage_setup):
        ic = tff.ImageFileCollection(triage_setup.test_dir)
        ic.keywords = []
        assert ['file'] == ic.keywords

    def test_header_and_filename(self, triage_setup):
        collection = tff.ImageFileCollection(location=triage_setup.test_dir,
                                             keywords=['imagetyp'])
        for header, fname in collection.headers(return_fname=True):
            assert (fname in collection._paths())
            assert (isinstance(header, fits.Header))

    def test_dir_with_no_fits_files(self, tmpdir):
        empty_dir = tmpdir.mkdtemp()
        some_file = empty_dir.join('some_file.txt')
        some_file.dump('words')
        print(empty_dir.listdir())
        collection = tff.ImageFileCollection(location=empty_dir.strpath,
                                             keywords=['imagetyp'])
        assert (collection.summary_info is None)
        for hdr in collection.headers():
            # this statement should not be reached if there are no FITS files
            assert 0

    def test_dir_with_no_keys(self, tmpdir, caplog):
        # This test should fail if the FITS files in the directory
        # are actually read.
        bad_dir = tmpdir.mkdtemp()
        not_really_fits = bad_dir.join('not_fits.fit')
        not_really_fits.dump('I am not really a FITS file')
        # make sure an error will be generated if the FITS file is read
        with pytest.raises(IOError):
            fits.getheader(not_really_fits.strpath)
        ic = tff.ImageFileCollection(location=bad_dir.strpath)
        # ImageFileCollection will suppress the IOError but log a warning
        # so check the log for the appropriate warning
        warnings = [rec for rec in caplog.records()
                    if ((rec.levelno == logging.WARN) &
                        ('Unable to get FITS header' in rec.message))]
        assert (len(warnings) == 0)

    def test_fits_summary_when_keywords_are_not_subset(self, triage_setup):
        """
        Catch case when there is overlap between keyword list
        passed to the ImageFileCollection and to files_filtered
        but the latter is not a subset of the former.
        """
        ic = tff.ImageFileCollection(triage_setup.test_dir,
                                     keywords=['imagetyp', 'exptime'])
        n_files = len(ic.files)
        files_missing_this_key = ic.files_filtered(imagetyp='*',
                                                   monkeys=None)
        assert(n_files > 0)
        assert(n_files == len(files_missing_this_key))

    def test_duplicate_keywords_in_setting(self, triage_setup):
        keywords_in = ['imagetyp', 'a', 'a']
        ic = tff.ImageFileCollection(triage_setup.test_dir,
                                     keywords=keywords_in)
        for key in set(keywords_in):
            assert (key in ic.keywords)
        # one keyword gets added: file
        assert len(ic.keywords) < len(keywords_in) + 1

    def test_keyword_includes_file(self, triage_setup):
        keywords_in = ['file', 'imagetyp']
        ic = tff.ImageFileCollection(triage_setup.test_dir,
                                     keywords=keywords_in)
        assert 'file' in ic.keywords
        file_keywords = [key for key in ic.keywords if key == 'file']
        assert len(file_keywords) == 1

    def test_setting_keywords_to_none(self, triage_setup):
        ic = tff.ImageFileCollection(triage_setup.test_dir,
                                     keywords=['imagetyp'])
        ic.keywords = None
        assert ic.summary_info == []

    def test_getting_value_for_keyword(self, triage_setup):
        ic = tff.ImageFileCollection(triage_setup.test_dir,
                                     keywords=['imagetyp'])
        # Does it fail if the keyword is not in the summary?
        with pytest.raises(ValueError):
            ic.values('filter')
        # If I ask for unique values do I get them?
        values = ic.values('imagetyp', unique=True)
        print(ic.summary_info['imagetyp'])
        assert values == list(set(ic.summary_info['imagetyp']))
        assert len(values) < len(ic.summary_info['imagetyp'])
        # Does the list of non-unique values match the raw column?
        values = ic.values('imagetyp', unique=False)
        assert values == list(ic.summary_info['imagetyp'])
        # Does unique actually default to false?
        values2 = ic.values('imagetyp')
        assert values == values2

    def test_collection_when_one_file_not_fits(self, triage_setup):
        not_fits = 'foo.fit'
        path_bad = os.path.join(triage_setup.test_dir, not_fits)
        # create an empty file...
        with open(path_bad, 'w'):
            pass
        ic = tff.ImageFileCollection(triage_setup.test_dir,
                                     keywords=['imagetyp'])
        assert not_fits not in ic.summary_info['file']
        os.remove(path_bad)

    def test_data_type_mismatch_in_fits_keyword_values(self, tmpdir,
                                                       triage_setup):
        # If one keyword has an unexpected type, do we notice?
        img = np.uint16(np.arange(100))
        bad_filter = fits.PrimaryHDU(img)
        bad_filter.header['imagetyp'] = IRAF_image_type('light')
        bad_filter.header['filter'] = 15.0
        path_bad = os.path.join(triage_setup.test_dir, 'bad_filter.fit')
        bad_filter.writeto(path_bad)
        ic = tff.ImageFileCollection(triage_setup.test_dir,
                                     keywords=['filter'])
        # dtype is object when there is a mix of types
        assert ic.summary_info['filter'].dtype == np.dtype('O')
        os.remove(path_bad)

    def test_filter_by_numerical_value(self, triage_setup):
        ic = tff.ImageFileCollection(triage_setup.test_dir, keywords=['naxis'])
        should_be_zero = ic.files_filtered(naxis=2)
        assert len(should_be_zero) == 0
        should_not_be_zero = ic.files_filtered(naxis=1)
        assert len(should_not_be_zero) == triage_setup.n_test['files']

    def test_unknown_generator_type_raises_error(self, triage_setup):
        ic = tff.ImageFileCollection(triage_setup.test_dir, keywords=['naxis'])
        with pytest.raises(ValueError):
            for foo in ic._generator('not a real generator'):
                pass

    def test_setting_write_location_to_bad_dest_raises_error(self, tmpdir,
                                                             triage_setup):
        new_tmp = tmpdir.mkdtemp()
        os.chmod(new_tmp.strpath, stat.S_IREAD)
        ic = tff.ImageFileCollection(triage_setup.test_dir, keywords=['naxis'])
        with pytest.raises(IOError):
            for hdr in ic.headers(save_location=new_tmp.strpath):
                pass

    def test_initializing_from_table(self, triage_setup):
        keys = ['imagetyp', 'filter']
        ic = tff.ImageFileCollection(triage_setup.test_dir,
                                     keywords=keys)
        table = ic.summary_info
        table_path = os.path.join(triage_setup.test_dir, 'input_tbl.csv')
        nonsense = 'forks'
        table['imagetyp'][0] = nonsense
        table.write(table_path, format='ascii', delimiter=',')
        ic = tff.ImageFileCollection(location=None, info_file=table_path)
        # keywords can only have been set from saved table
        for key in keys:
            assert key in ic.keywords
        # no location, so should be no files
        assert len(ic.files) == 0
        # no location, so no way to iterate over files
        with pytest.raises((AttributeError, TypeError)):
            for h in ic.headers():
                pass
        ic = tff.ImageFileCollection(location=triage_setup.test_dir,
                                     info_file=table_path)
        # we now have a location, so did we get files?
        assert len(ic.files) == len(table)
        # Is the summary table masked?
        assert ic.summary_info.masked
        # can I loop over headers?
        for h in ic.headers():
            assert isinstance(h, fits.Header)
        # Does ImageFileCollection summary contain values from table?
        assert nonsense in ic.summary_info['imagetyp']

    def test_initializing_from_table_file_that_does_not_exist(self,
                                                              triage_setup,
                                                              caplog):
        # Do we get a warning if we try reading a file that doesn't exist,
        # but where we can initialize from a directory?
        ic = tff.ImageFileCollection(location=triage_setup.test_dir,
                                     info_file='iufadsdhfasdifre')
        warnings = [rec for rec in caplog.records()
                    if ((rec.levelno == logging.WARN) &
                        ('Unable to open table file' in rec.message))]
        assert (len(warnings) == 1)
        # Do we raise an error if the table name is bad AND the location
        # is None?
        with pytest.raises(IOError):
            ic = tff.ImageFileCollection(location=None,
                                         info_file='iufadsdhfasdifre')
        # Do we raise an error if the table name is bad AND
        # the location is given but is bad?
        with pytest.raises(OSError):
            ic = tff.ImageFileCollection(location='dasifjoaurun',
                                         info_file='iufadsdhfasdifre')

    def test_initialization_with_no_keywords(self, triage_setup):
        ic = tff.ImageFileCollection(location=triage_setup.test_dir)
        # iteration below failed before bugfix...
        execs = 0
        for h in ic.headers():
            execs += 1
            print(h)
        assert not execs

    def check_all_keywords_in_collection(self, image_collection):
        lower_case_columns = [c.lower() for c in
                              image_collection.summary_info.colnames]
        for h in image_collection.headers():
            for k in h:
                assert k.lower() in lower_case_columns

    def test_tabulate_all_keywords(self, triage_setup):
        ic = tff.ImageFileCollection(location=triage_setup.test_dir,
                                     keywords='*')
        self.check_all_keywords_in_collection(ic)

    def test_summary_table_is_always_masked(self, triage_setup):
        # First, try grabbing all of the keywords
        ic = tff.ImageFileCollection(location=triage_setup.test_dir,
                                     keywords='*')
        assert ic.summary_info.masked
        # Now, try keywords that every file will have
        ic.keywords = ['bitpix']
        assert ic.summary_info.masked
        # What about keywords that include some that will surely be missing?
        ic.keywords = ['bitpix', 'dsafui']
        assert ic.summary_info.masked

    def test_case_of_keywords_respected(self, triage_setup):
        keywords_in = ['BitPix', 'instrume', 'NAXIS']
        ic = tff.ImageFileCollection(location=triage_setup.test_dir,
                                     keywords=keywords_in)
        for key in keywords_in:
            assert key in ic.summary_info.colnames

    def test_grabbing_all_keywords_and_specific_keywords(self, triage_setup):
        keyword_not_in_headers = 'OIdn89!@'
        ic = tff.ImageFileCollection(triage_setup.test_dir,
                                     keywords=['*', keyword_not_in_headers])
        assert keyword_not_in_headers in ic.summary_info.colnames
        self.check_all_keywords_in_collection(ic)

    def test_grabbing_all_keywords_excludes_empty_key(self, triage_setup):
        # This test needs a file with a blank keyword in it to ensure
        # that case is handled correctly.
        blank_keyword = fits.PrimaryHDU()
        blank_keyword.data = np.zeros((100, 100))
        blank_keyword.header[''] = 'blank'

        blank_keyword.writeto(os.path.join(triage_setup.test_dir,
                                           'blank.fits'))

        ic = tff.ImageFileCollection(triage_setup.test_dir, keywords='*')
        print(ic.summary_info.colnames)
        assert 'col0' not in ic.summary_info.colnames

    def test_header_with_long_history_roundtrips_to_disk(self, triage_setup):
        # I tried combing several history comments into one table entry with
        # '\n'.join(history), which resulted in a table that couldn't
        # round trip to disk because on read the newline character was
        # interpreted as...a new line! This test is a check against future
        # foolishness.
        from astropy.table import Table
        img = np.uint16(np.arange(100))
        long_history = fits.PrimaryHDU(img)
        long_history.header['imagetyp'] = IRAF_image_type('bias')
        long_history.header['history'] = 'Something happened'
        long_history.header['history'] = 'Then something else happened'
        long_history.header['history'] = 'And then something odd happened'
        path_history = os.path.join(triage_setup.test_dir, 'long_history.fit')
        long_history.writeto(path_history)
        ic = tff.ImageFileCollection(triage_setup.test_dir, keywords='*')
        ic.summary_info.write('test_table.txt', format='ascii.csv')
        table_disk = Table.read('test_table.txt', format='ascii.csv')
        assert len(table_disk) == len(ic.summary_info)

    def test_refresh_method_sees_added_keywords(self, triage_setup):
        ic = tff.ImageFileCollection(triage_setup.test_dir, keywords='*')
        # Add a keyword I know isn't already in the header to each file.
        not_in_header = 'BARKARK'
        for h in ic.headers(overwrite=True):
            h[not_in_header] = True
        print(h)
        assert not_in_header not in ic.summary_info.colnames

        ic.refresh()
        # After refreshing the odd keyword should be present.
        print(ic.keywords)
        assert not_in_header.lower() in ic.summary_info.colnames

    def test_refresh_method_sees_added_files(self, triage_setup):
        ic = tff.ImageFileCollection(triage_setup.test_dir, keywords='*')
        # Compressed files don't get copied. Not sure why...
        original_len = len(ic.summary_info) - triage_setup.n_test['compressed']
        # Generate additional files in this directory
        for h in ic.headers(save_with_name="_foo"):
            pass
        ic.refresh()
        new_len = len(ic.summary_info) - triage_setup.n_test['compressed']
        print(ic.summary_info['file'])
        assert new_len == 2 * original_len
